import configparser
import pathlib


class ExaLocalConfig(object):
    """
    Parse local config file
    Prepare arguments for pyexasol.connect() method

    Arguments accepting functions and class names are not supported
    You may pass such arguments directly in code using **kwargs
    """

    default_filename = '.pyexasol.ini'

    arg_types = {
        'autocommit': bool,
        'socket_timeout': int,
        'query_timeout': int,
        'compression': bool,
        'encryption': bool,
        'fetch_dict': bool,
        'fetch_size_bytes': int,
        'lower_ident': bool,
        'verbose_error': bool,
        'debug': bool,
        'udf_output_port': int,
    }

    def __init__(self, config_path=None):
        if config_path:
            self.path = pathlib.Path(config_path)
        else:
            self.path = self.get_default_path()

        if not self.path.exists():
            raise RuntimeError(f"PyEXASOL config file [{self.path}] not found")

        self.parser = configparser.ConfigParser()
        self.parser.read(self.path)

    def get_args(self, section):
        args = dict()

        if not self.parser.has_section(section):
            raise ValueError(f'Section [{section}] in PyEXASOL config file [{self.path}] not found')

        for k in self.parser[section]:
            t = self.get_arg_type(k)

            if t == bool:
                args[k] = self.parser[section].getboolean(k)
            elif t == int:
                args[k] = self.parser[section].getint(k)
            elif t == float:
                args[k] = self.parser[section].getfloat(k)
            else:
                args[k] = self.parser[section].get(k)

        return args

    def get_default_path(self):
        return pathlib.Path.home() / self.default_filename

    def get_arg_type(self, k):
        return self.arg_types.get(k, str)



"""
TCP Server to capture UDF script output.
Unlike HTTP Transport, we have to use real server with proper bindings.

Incoming connections may be blocked by firewalls!
It mostly affects local laptops trying to connect to Exasol located in remote data center.

This module can work in two modes:

    1) DEBUG MODE
    Useful for manual debugging during UDF script development.
    Accepts connections from all VM's, but displays output of first connected VM only.
    Runs forever, until stopped by Ctrl + C (SIGTERM).

    How to run: python -m pyexasol script_debug


    2) SCRIPT MODE
    Useful for production usage and during last stages of development.
    Accepts connections from all VM's and stores output into separate log files.
    Runs for one SQL statement only, stops automatically.

    How to run: ExaConnection.execute_with_udf_output()


We use ThreadingMixIn because:
a) Workload is pure I/O, so GIL should not be a problem.
b) Potential amount of VM's connected in parallel is huge and may surpass 1000+, which may cause problems with forks.

"""
import socket
import socketserver
import sys
import os
import shutil
import multiprocessing
import pathlib


class ExaScriptOutput(object):
    def __init__(self, server_host=None, server_port=None):
        self.server_host = str(server_host) if server_host else '0.0.0.0'
        self.server_port = int(server_port) if server_port else 0

        self.server = None
        self.proc = None

    def init_debug_mode(self):
        self.server = ExaScriptOutputServer((self.server_host, self.server_port), ExaScriptOutputDebugModeHandler)
        return self.server.get_output_address()

    def wait_debug_mode(self):
        # Stop server with Ctrl + C
        try:
            self.server.serve_forever()
        except KeyboardInterrupt:
            pass

        self.server.server_close()

    def init_script_mode(self, output_dir):
        output_dir = pathlib.Path(output_dir)

        if not output_dir.is_dir():
            raise ValueError(f"Output_dir does not exist or not a directory: {output_dir}")

        self.proc = ExaScriptOutputProcess(self.server_host, self.server_port, output_dir)
        self.proc.start()

        output_address = self.proc.read_pipe.recv()

        self.proc.write_pipe.close()
        self.proc.read_pipe.close()

        return output_address

    def wait_script_mode(self):
        self.proc.join()

        if self.proc.exitcode != 0:
            raise RuntimeError(f"Output server process was terminated with exitcode: {self.proc.exitcode}")

    def terminate_script_mode(self):
        self.proc.terminate()


class ExaScriptOutputProcess(multiprocessing.Process):
    def __init__(self, server_host, server_port, output_dir):
        self.server_host = server_host
        self.server_port = server_port
        self.output_dir = output_dir

        self.read_pipe, self.write_pipe = multiprocessing.Pipe(False)
        super().__init__()

    def run(self):
        self.read_pipe.close()

        server = ExaScriptOutputServer((self.server_host, self.server_port), ExaScriptOutputScriptModeHandler)
        server.output_dir = self.output_dir

        self.write_pipe.send(server.get_output_address())
        self.write_pipe.close()

        # Stopped inside handler, after closing last connection
        server.serve_forever()
        server.server_close()


class ExaScriptOutputServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    connected_clients = 0
    total_clients = 0

    # Stop all sub-threads immediately
    daemon_threads = True
    allow_reuse_address = True

    def get_output_address(self):
        return f"{socket.getfqdn()}:{self.socket.getsockname()[1]}"


class ExaScriptOutputHandler(socketserver.StreamRequestHandler):
    def setup(self):
        super().setup()
        self.server.connected_clients += 1
        self.server.total_clients += 1

    def finish(self):
        super().finish()
        self.server.connected_clients -= 1


class ExaScriptOutputDebugModeHandler(ExaScriptOutputHandler):
    def handle(self):
        if self.server.connected_clients == 1:
            print('\n-------- NEW STATEMENT --------', flush=True)
            # Read and flush line-by-line, show log to user as soon as possible

            for line in self.rfile:
                sys.stdout.buffer.write(line)
                sys.stdout.buffer.flush()
        else:
            dst = open(os.devnull, 'wb')
            shutil.copyfileobj(self.rfile, dst)
            dst.close()


class ExaScriptOutputScriptModeHandler(ExaScriptOutputHandler):
    def handle(self):
        path = self.server.output_dir / (str(self.server.total_clients).rjust(5, '0') + '.log')
        dst = open(path, 'w+b')

        shutil.copyfileobj(self.rfile, dst)
        dst.close()

    def finish(self):
        super().finish()

        if self.server.connected_clients == 0:
            self.server.shutdown()



"""
Extension with Exasol-specific helper functions
"""


class ExaExtension(object):
    def __init__(self, connection):
        self.connection = connection
        self.reserved_words = None

    def get_columns(self, object_name):
        """
        Get information about columns of table or view (Websocket format)
        Object name may be passed as tuple to specify custom schema
        """
        return self.get_columns_sql("SELECT * FROM {object_name!i}", {'object_name': object_name})

    def get_columns_sql(self, query, query_params=None):
        """
        Get columns of SQL query without executing it (Websocket format)
        Relies on prepared statement which are closed immediately without execution
        """
        st = self.connection._statement(query, query_params)
        st._prepare()

        columns = st.columns()
        st.close()

        return columns

    def get_sys_columns(self, object_name):
        """
        Get information about columns of table or view (SYS format)
        Object name may be passed as tuple to specify custom schema
        """
        if isinstance(object_name, tuple):
            schema = str(object_name[0]).upper()
            object_name = str(object_name[1]).upper()
        else:
            schema = self.connection.current_schema()
            object_name = str(object_name).upper()

        sql = """
            SELECT c.column_name, c.column_type, c.column_maxsize, c.column_num_scale,
                   c.column_is_nullable, c.column_is_distribution_key, c.column_default,
                   c.column_comment, t.type_name
            FROM EXA_ALL_COLUMNS c
                JOIN EXA_SQL_TYPES t ON (c.column_type_id=t.type_id)
            WHERE c.column_schema={schema}
                AND c.column_table={object_name}
            ORDER BY c.column_ordinal_position
        """

        st = self._execute(sql, {'schema': schema, 'object_name': object_name})
        res = list()

        for r in st:
            res.append({
                'name': r['column_name'],
                'type': r['type_name'],
                'sql_type': r['column_type'],
                'size': r['column_maxsize'],
                'scale': r['column_num_scale'],
                'nulls': r['column_is_nullable'],
                'distribution_key': r['column_is_distribution_key'],
                'default': r['column_default'],
                'comment': r['column_comment'],
            })

        return res

    def get_sys_tables(self, schema=None, table_name_prefix=''):
        """
        Get information about tables in selected schema(SYS format)
        Output may be optionally filtered by table name prefix
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = str(schema).upper()

        table_name_prefix = self.connection.format.escape_like(table_name_prefix).upper()

        sql = """
            SELECT *
            FROM EXA_ALL_TABLES
            WHERE table_schema={schema}
                AND table_name LIKE '{table_name_prefix!r}%'
            ORDER BY table_name ASC
        """

        st = self._execute(sql, {'schema': schema, 'table_name_prefix': table_name_prefix})
        res = list()

        for r in st:
            res.append({
                'table_name': r['table_name'].lower() if self.connection.lower_ident else r['table_name'],
                'table_schema': r['table_schema'].lower() if self.connection.lower_ident else r['table_schema'],
                'table_is_virtual': r['table_is_virtual'],
                'table_has_distribution_key': r['table_has_distribution_key'],
                'table_comment': r['table_comment'],
            })

        return res

    def get_sys_views(self, schema=None, view_name_prefix=''):
        """
        Get information about views in selected schema(SYS format)
        Output may be optionally filtered by view name prefix
        """
        if schema is None:
            schema = self.connection.current_schema()
        else:
            schema = str(schema).upper()

        view_name_prefix = self.connection.format.escape_like(view_name_prefix).upper()

        sql = """
            SELECT *
            FROM EXA_ALL_VIEWS
            WHERE view_schema={schema}
                AND view_name LIKE '{view_name_prefix!r}%'
            ORDER BY view_name ASC
        """

        st = self._execute(sql, {'schema': schema, 'view_name_prefix': view_name_prefix})
        res = list()

        for r in st:
            res.append({
                'view_name': r['view_name'].lower() if self.connection.lower_ident else r['view_name'],
                'view_schema': r['view_schema'].lower() if self.connection.lower_ident else r['view_schema'],
                'scope_schema': r['scope_schema'].lower() if self.connection.lower_ident else r['scope_schema'],
                'view_text': r['view_text'],
                'view_comment': r['view_comment'],
            })

        return res

    def get_sys_schemas(self, schema_name_prefix=''):
        """
        Get information about schemas (SYS format)
        Output may be optionally filtered by schema name prefix
        """
        schema_name_prefix = self.connection.format.escape_like(schema_name_prefix).upper()

        sql = """
            SELECT *
            FROM EXA_SCHEMAS
            WHERE schema_name LIKE '{schema_name_prefix!r}%'
            ORDER BY schema_name ASC
        """

        st = self._execute(sql, {'schema_name_prefix': schema_name_prefix})
        res = list()

        for r in st:
            res.append({
                'schema_name': r['schema_name'].lower() if self.connection.lower_ident else r['schema_name'],
                'schema_owner': r['schema_owner'].lower() if self.connection.lower_ident else r['schema_owner'],
                'schema_is_virtual': r['schema_is_virtual'],
                'schema_comment': r['schema_comment'],
            })

        return res

    def get_reserved_words(self):
        """
        Get reserved keywords which cannot be used as identifiers without double-quotes
        Never hard-code this list! It changes with every Exasol versions
        """
        if self.reserved_words is None:
            sql = """
                SELECT keyword
                FROM EXA_SQL_KEYWORDS
                WHERE reserved IS TRUE
                ORDER BY keyword
            """

            self.reserved_words = self._execute(sql).fetchcol()

        return self.reserved_words

    def get_disk_space_usage(self):
        """
        Exasol still lacks standard function to measure actual disk space usage
        We're trying to mitigate this problem by making custom function
        """
        sql = """
            SELECT measure_time,
                   (committed_size * redundancy + temp_swap_data) AS occupied_size,
                   (device_size * redundancy + hdd_free) AS total_size
            FROM "$EXA_STATS_DB_SIZE"
            ORDER BY measure_time DESC
            LIMIT 1
        """

        row = self._execute(sql).fetchone()

        if row is None:
            return None

        row['occupied_size'] = int(row['occupied_size'])
        row['total_size'] = int(row['total_size'])

        row['free_size'] = row['total_size'] - row['occupied_size']
        row['occupied_size_percent'] = round(row['occupied_size'] / row['total_size'] * 100, 2)

        return row

    def export_to_pandas_with_dtype(self, query_or_table, query_params=None):
        """
        Export to pandas and attempt to guess correct dtypes based on Exasol columns
        Since pandas has significantly smaller range of allowed values, this function makes many assumptions
        Please use it as baseline for your own function for complex cases

        Small decimal -> int32
        Big decimal -> int64
        Double -> float64
        Date, Timestamp -> datetime64[ns]
        Everything else -> category (!)
        """

        if query_params:
            query_or_table = self.connection.format.format(query_or_table, **query_params)

        if str(query_or_table).strip().find(' ') == -1:
            columns = self.get_columns(query_or_table)
        else:
            columns = self.get_columns_sql(query_or_table)

        params = {
            'names': list(),
            'dtype': dict(),
            'parse_dates': list(),
            'na_values': dict(),
            'infer_datetime_format': True,
            'engine': 'c',
            'skip_blank_lines': False,
        }

        for k, c in columns.items():
            params['names'].append(k)

            if c['type'] == 'DATE':
                params['dtype'][k] = 'object'
                params['na_values'][k] = '0001-01-01'
                params['parse_dates'].append(k)

            elif c['type'] == 'TIMESTAMP':
                params['dtype'][k] = 'object'
                params['na_values'][k] = '0001-01-01 00:00:00'
                params['parse_dates'].append(k)

            elif c['type'] == 'DECIMAL':
                if c['scale'] > 0:
                    params['dtype'][k] = 'float64'
                else:
                    if c['precision'] <= 9:
                        params['dtype'][k] = 'int32'
                    else:
                        params['dtype'][k] = 'int64'

            elif c['type'] == 'DOUBLE':
                params['dtype'][k] = 'float64'

            else:
                params['dtype'][k] = 'category'

        def callback(pipe, dst, **kwargs):
            import pandas
            return pandas.read_csv(pipe, **kwargs)

        return self.connection.export_to_callback(callback, None, query_or_table, None, params)

    def _execute(self, query, query_params=None):
        st = self.connection._statement(query, query_params)

        st.fetch_dict = True
        st.fetch_mapper = None
        st.lower_ident = True

        st._execute()

        return st



DEFAULT_PORT = 8563
DEFAULT_AUTOCOMMIT = True

DEFAULT_SOCKET_TIMEOUT = 10
DEFAULT_QUERY_TIMEOUT = 0

DEFAULT_FETCHMANY_SIZE = 10000
DEFAULT_FETCH_SIZE_BYTES = 5 * 1024 * 1024

DRIVER_NAME = 'PyEXASOL'

LOGGER_MESSAGE_TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
LOGGER_FILENAME_TIMESTAMP_FORMAT = '%Y%m%d_%H%M%S_%f'
LOGGER_MAX_JSON_LENGTH = 50000



"""
Standalone functions to be executed as python module
Usage: python -m pyexasol <command> <options>
"""

import argparse
from .script_output import ExaScriptOutput
from .version import __version__


parser = argparse.ArgumentParser(prog=f'python -m pyexasol')
subparsers = parser.add_subparsers(title=None, dest='command', metavar='<command>')
subparsers.required = True

subp = subparsers.add_parser('script_debug', help='Run script output server for debugging')
subp.add_argument('--host', default=None, help='Specific address to bind TCPServer (default: 0.0.0.0)')
subp.add_argument('--port', default=None, help='Specific port to bind TCPServer (default: random port)')

subp = subparsers.add_parser('version', help='Show PyEXASOL version')

args = parser.parse_args()

if args.command == 'script_debug':
    obj = ExaScriptOutput(args.host, args.port)

    output_address = obj.init_debug_mode()
    print(f"ALTER SESSION SET SCRIPT_OUTPUT_ADDRESS = '{output_address}';", flush=True)

    obj.wait_debug_mode()
elif args.command == 'version':
    print(f'PyEXASOL {__version__}')




__all__ = [
    'connect',
    'exasol_mapper',
    'ExaError',
    'ExaCommunicationError',
    'ExaRuntimeError',
    'ExaRequestError',
    'ExaQueryError',
    'ExaQueryTimeoutError',
    'ExaConnection',
    'ExaStatement',
    'ExaFormatter',
    'ExaLogger',
    'ExaExtension',
    'ExaHTTPTransportWrapper',
    'ExaScriptOutput',
    'ExaLocalConfig',
    'HTTP_EXPORT',
    'HTTP_IMPORT',
]

from .version import __version__
from .exceptions import ExaError, ExaCommunicationError, ExaRuntimeError, ExaRequestError, \
                        ExaQueryError, ExaQueryTimeoutError
from .connection import ExaConnection
from .statement import ExaStatement
from .formatter import ExaFormatter
from .local_config import ExaLocalConfig
from .logger import ExaLogger
from .ext import ExaExtension
from .mapper import exasol_mapper
from .http_transport import ExaHTTPTransportWrapper, HTTP_EXPORT, HTTP_IMPORT
from .script_output import ExaScriptOutput


def connect(**kwargs) -> ExaConnection:
    """
    Constructor of connection objects
    Please see ExaConnection object for list of arguments
    """
    if 'cls_connection' in kwargs:
        connection_cls = kwargs['cls_connection']

        if not issubclass(connection_cls, ExaConnection):
            raise ValueError(f"Class [{kwargs['cls_connection']}] is not subclass of ExaConnection")
    else:
        connection_cls = ExaConnection

    if 'cls_statement' in kwargs and not issubclass(kwargs['cls_statement'], ExaStatement):
        raise ValueError(f"Class [{kwargs['cls_statement']}] is not subclass of ExaStatement")

    if 'cls_formatter' in kwargs and not issubclass(kwargs['cls_formatter'], ExaFormatter):
        raise ValueError(f"Class [{kwargs['cls_formatter']}] is not subclass of ExaFormatter")

    if 'cls_logger' in kwargs and not issubclass(kwargs['cls_logger'], ExaLogger):
        raise ValueError(f"Class [{kwargs['cls_logger']}] is not subclass of ExaLogger")

    if 'cls_extension' in kwargs and not issubclass(kwargs['cls_extension'], ExaExtension):
        raise ValueError(f"Class [{kwargs['cls_extension']}] is not subclass of ExaExtension")

    return connection_cls(**kwargs)


def connect_local_config(config_section, config_path=None, cls_local_config=ExaLocalConfig, **kwargs) -> ExaConnection:
    """
    Constructor of connection objects based on local config file
    Default config path is ~/.pyexasol.ini

    Extra arguments override values from config

    :param config_section: Name of config section (required!)
    :param config_path: Custom path to local config file
    :param cls_local_config: Overloaded ExaLocalConfig class
    :param kwargs: Arguments for "connect()" function
    """
    if not issubclass(cls_local_config, ExaLocalConfig):
        raise ValueError(f"Class [{cls_local_config}] is not subclass of ExaLocalConfig")

    conf = cls_local_config(config_path)
    conf_args = conf.get_args(config_section)

    return connect(**{**conf_args, **kwargs})


def http_transport(dsn, mode, compression=False, encryption=False) -> ExaHTTPTransportWrapper:
    """
    Constructor of HTTP Transport wrapper objects for IMPORT / EXPORT parallelism

    How to use this:
    1) Parent process opens main connection to Exasol with .connect()
    2) Parent process creates any number of child processes (possibly on remote host)
    3) Every child process inits HTTP transport sub-connection with .http_transport()
           and gets proxy "host:port" string using .get_proxy()
    4) Every child process sends proxy string to parent process using any communication method
    5) Parent process runs .export_parallel(), which executes EXPORT query in Exasol
    6) Every child process receives chunk of data, process it and finish
    7) Parent process waits for EXPORT query and child processes to finish

    All child processes should run in parallel. It is not possible to run some processes first, than run some more.
    """
    return ExaHTTPTransportWrapper(dsn, mode, compression, encryption)
